{
 A common file for RVMedia demos: camera list loader.
}

unit CamList;

{$IFDEF FPC}
{$MODE Delphi}
{$ENDIF}
{.$DEFINE CAMLIST_SAVELOG}

interface

uses SampleCamList, Classes, StdCtrls, MRVCamera, SysUtils, Forms, MRVType;

const
  STS_PLAY       = 'Preparing to receive video...';
  STS_PLAYING    = 'Playing';
  STS_DONEVIDEO  = 'Video stream is finished';
  STS_ERRORVIDEO = 'Video stream error';
  STS_PRESENT    = 'Camera exists';
  STS_NOTFOUND   = 'Camera is not found';
  STS_SEARCH     = 'Connecting to the camera...';
  STS_RUNNING    = 'Playing video stream...';
  STS_STOPPING   = 'Stopping video stream...';

type
  TStatusChangeEvent = procedure(Camera: TRVCamera; const Status: string)
    of object;
  TEnableControlsEvent = procedure(Enabled: Boolean) of object;

  TCamStruct = record
    Alias: string;
    URL: string;
    Host: string;
    Port: word;
    UserName: string;
    UserPassword: string;
    VideoImagesURLs: string;
  end;

  PCamStruct = ^TCamStruct;

  TCamListLoader = class
  private
    FList:           TList;
    FUSBDeviceCount: Integer;

    procedure SetStatus(Camera: TRVCamera; Status: String);
    procedure ClearList;
  public
    DevicePath : TStringList;
    OnStatusChange:     TStatusChangeEvent;
    OnEnableControls:   TEnableControlsEvent;
    OnStartVideoStream: TNotifyEvent;
    OnEndVideoStream:   TRVCamDoneEvent;
    OnVideoStart:       TNotifyEvent;
    constructor Create;
    destructor Destroy; override;
    procedure Load(CamList: TStrings; OnlyControllable: Boolean = False);
    procedure RunCamera(RVCamera: TRVCamera);
    procedure PlayFile(RVCamera: TRVCamera; const Filename: String);
    procedure Select(Index: Integer; RVCamera: TRVCamera;IniciaCamera:Boolean=True);
    procedure SelectDevicePath(aDevicePath: String; RVCamera: TRVCamera;IniciaCamera:Boolean=True);
    procedure SelectIPCamera(RVCamera: TRVCamera; Index: Integer);
    procedure SearchComplete(RVCamera: TRVCamera; Status: Integer;
      IsSelectedCamera: Boolean = True);
    procedure DoOnEndVideoStream(Sender: TObject; Status: Integer);
    procedure DoOnVideoStart(Sender: TObject);
    procedure DoOnStartVideoStream(Sender: TObject);
  end;

var
  CamListLoader: TCamListLoader;

implementation

procedure AddTextLog(GUIDMy, Text: string);
{$IFDEF CAMLIST_SAVELOG}
var
  F:       TextFile;
  fn, Dir: string;
  {$ENDIF}
begin
  {$IFDEF CAMLIST_SAVELOG}
  if GUIDMy = '' then
    exit;

  Dir := ExtractFilePath(Application.ExeName) + '\debug_log\';
  CreateDir(Dir);
  fn := Dir + GUIDMy + '.camlist';
  AssignFile(F, fn);
  if FileExists(fn) then
    Append(F)
  else
    Rewrite(F);

  Writeln(F, Text);
  Flush(F);
  CloseFile(F);
  {$ENDIF}
end;

{------------------------------------------------------------------------------}
function GetValue(s: string): string;
var
  p: Integer;
begin
  p := Pos('=', s);
  Result := Trim(Copy(s, p + 1, length(s)));
end;

{------------------------------------------------------------------------------}
function GetParam(var s: string): string;
var
  p: Integer;
begin
  p := Pos('|', s);
  if p > 0 then
    Result := Trim(Copy(s, 1, p - 1))
  else
    Result := Trim(Copy(s, 1, length(s)));
  Delete(s, 1, p);
end;

{============================ TCamListLoader ==================================}
constructor TCamListLoader.Create;
begin
  inherited;
  FList := TList.Create;
  FUSBDeviceCount := 0;

  DevicePath := TStringList.Create;
end;

{------------------------------------------------------------------------------}
destructor TCamListLoader.Destroy;
begin
  ClearList;
  FList.Free;

  DevicePath.Clear;
  FreeAndNil(DevicePath);
  inherited;
end;

{------------------------------------------------------------------------------}
procedure TCamListLoader.ClearList;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    Dispose(PCamStruct(FList.Items[i]));
end;

{------------------------------------------------------------------------------}
// Loading a list of cameras in CamList (all cameras or only controllable cameras)
procedure TCamListLoader.Load(CamList: TStrings; OnlyControllable: Boolean);
var
  i:            Integer;
  tmpUSBWebCam: TRVCamera;
  C:            ^TCamStruct;
  s:            string;
begin
  ClearList;
  FList.Clear;

  for i := 0 to length(SampleCameraList) - 1 do
  begin
    New(C);
    s := SampleCameraList[i];
    C^.Alias := GetValue(GetParam(s));
    C^.URL := GetValue(GetParam(s));
    C^.Host := GetValue(GetParam(s));
    C^.Port := StrToIntDef(GetValue(GetParam(s)), 0);
    C^.UserName := GetValue(GetParam(s));
    C^.UserPassword := GetValue(GetParam(s));
    C^.VideoImagesURLs := GetValue(GetParam(s));
    if not OnlyControllable or (GetValue(GetParam(s)) = '1') then
      FList.Add(C);
  end;

  CamList.Clear;
  CamList.BeginUpdate;
  try
    if not OnlyControllable then
    begin
      tmpUSBWebCam := TRVCamera.Create(nil);
      try
        tmpUSBWebCam.DeviceType := rvdtWebCamera;
        tmpUSBWebCam.FillVideoDeviceList(CamList);
        FUSBDeviceCount := CamList.Count;

        // Rogrio
        for i := 0 to CamList.Count-1 do
          DevicePath.Add(tmpUSBWebCam.VideoDeviceIdList[i])

      finally
        tmpUSBWebCam.Free;
      end;
    end
    else
      FUSBDeviceCount := 0;
{    for i := 0 to FList.Count - 1 do
      CamList.Add(TCamStruct(FList[i]^).Alias);}
  finally
    CamList.EndUpdate;
  end;
end;

{------------------------------------------------------------------------------}
// Switching to the Index-th IP camera.
// If Index-1, initialting playing RVCamera without changing its properties
procedure TCamListLoader.SelectIPCamera(RVCamera: TRVCamera; Index: Integer);
var
  s, s0: string;
  C:     TCamStruct;
begin
  if RVCamera.Aborting then
    exit;

  if RVCamera.IsVideoPlaying then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist RVCamera.Abort;');
    RVCamera.Abort;
    AddTextLog(CAMERA_GUID, 'Camlist RVCamera.WaitForVideo;');
    RVCamera.WaitForVideo;
    if Application.Terminated then
      exit;
    AddTextLog(CAMERA_GUID, 'Camlist IsVideoPlaying finished');
  end;

  if Index >= 0 then
  begin
    C := TCamStruct(FList[Index]^);
    RVCamera.URL := C.URL;
    RVCamera.CameraHost := C.Host;
    RVCamera.CameraPort := C.Port;
    RVCamera.UserName := C.UserName;
    RVCamera.UserPassword := C.UserPassword;
    s := C.VideoImagesURLs;
    RVCamera.VideoImagesURLs.Clear;
    RVCamera.CycleVideoImages := s <> '';
    while s <> '' do
    begin
      if Pos(';', s) > 0 then
        s0 := Copy(s, 1, Pos(';', s) - 1)
      else
        s0 := s;
      Delete(s, 1, length(s0) + 1);
      RVCamera.VideoImagesURLs.Add(s0);
    end;
  end;

  AddTextLog(CAMERA_GUID, 'Camlist RVCamera.CameraHost ' + RVCamera.CameraHost);
  if RVCamera.CameraHost <> '' then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist RVCamera.SearchCamera');
    RVCamera.SearchCamera;
    AddTextLog(CAMERA_GUID, 'Camlist RVCamera.SetStatus STS_SEARCH');
    SetStatus(RVCamera, STS_SEARCH);
  end
  else
  begin
    AddTextLog(CAMERA_GUID, 'Camlist RunCamera');
    RunCamera(RVCamera);
  end;
  AddTextLog(CAMERA_GUID, 'Camlist SelectIPCamera finished');

  if Assigned(OnEnableControls) then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist OnEnableControls started');
    OnEnableControls(not RVCamera.Searching);
    AddTextLog(CAMERA_GUID, 'Camlist OnEnableControls finished');
  end;
end;

{------------------------------------------------------------------------------}
// Starting playing a video from the camera
procedure TCamListLoader.RunCamera(RVCamera: TRVCamera);
begin
  AddTextLog(CAMERA_GUID, 'Camlist RunCamera started');
  if RVCamera.IsVideoPlaying then
  begin
    AddTextLog(CAMERA_GUID,
      'Camlist RunCamera SetStatus(RVCamera, STS_STOPPING);');
    SetStatus(RVCamera, STS_STOPPING);
    AddTextLog(CAMERA_GUID, 'Camlist RunCamera RVCamera.Abort;');
    RVCamera.Abort;
    AddTextLog(CAMERA_GUID, 'Camlist RunCamera RVCamera.WaitForVideo;');
    RVCamera.WaitForVideo;
    AddTextLog(CAMERA_GUID, 'Camlist RunCamera Terminated');
    if Application.Terminated then
      exit;
  end;
  AddTextLog(CAMERA_GUID,
    'Camlist RunCamera SetStatus(RVCamera, STS_RUNNING);');
  SetStatus(RVCamera, STS_RUNNING);
  AddTextLog(CAMERA_GUID, 'Camlist RunCamera PlayVideoStream;');
  RVCamera.PlayVideoStream;
  AddTextLog(CAMERA_GUID, 'Camlist RunCamera finished');
end;

{------------------------------------------------------------------------------}
// Starting playing a video from the file
procedure TCamListLoader.PlayFile(RVCamera: TRVCamera; const Filename: String);
begin
  If RVCamera.IsVideoPlaying then
  begin
    SetStatus(RVCamera, STS_STOPPING);
    RVCamera.Abort;
    RVCamera.WaitForVideo;
    if Application.Terminated then
      exit;
  end;
  SetStatus(RVCamera, STS_RUNNING);
  RVCamera.PlayVideoFile(Filename);
end;

{------------------------------------------------------------------------------}
// Switch to the Index-th camera
procedure TCamListLoader.Select(Index: Integer; RVCamera: TRVCamera;iniciaCamera:Boolean);
begin
  AddTextLog(CAMERA_GUID, 'Camlist Select started');
  if Index < FUSBDeviceCount then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist Select DeviceType := rvdtWebCamera');
    RVCamera.DeviceType := rvdtWebCamera;
    AddTextLog(CAMERA_GUID, 'Camlist Select VideoDeviceIndex');
    RVCamera.VideoDeviceIndex := Index;
    AddTextLog(CAMERA_GUID, 'Camlist Select RunCamera');
    if IniciaCamera then
      RunCamera(RVCamera);
  end
  else
  begin
    AddTextLog(CAMERA_GUID,
      'Camlist Select RVCamera.DeviceType := rvdtIPCamera;');
    RVCamera.DeviceType := rvdtIPCamera;
    AddTextLog(CAMERA_GUID, 'Camlist Select SelectIPCamera');
    SelectIPCamera(RVCamera, Index - FUSBDeviceCount);
  end;
  AddTextLog(CAMERA_GUID, 'Camlist Select finished');
end;

procedure TCamListLoader.SelectDevicePath(aDevicePath: String; RVCamera: TRVCamera;IniciaCamera:Boolean=True);
var
  Index : Integer;
begin
  AddTextLog(CAMERA_GUID, 'Camlist SelectDevicePath started');
  Index := DevicePath.IndexOf(aDevicePath);
  Select(Index,RVCamera,IniciaCamera);
  AddTextLog(CAMERA_GUID, 'Camlist SelectDevicePath finished');
end;
procedure TCamListLoader.SearchComplete(RVCamera: TRVCamera; Status: Integer;
  IsSelectedCamera: Boolean);
begin
  AddTextLog(CAMERA_GUID, 'Camlist SearchComplete started');
  if IsSelectedCamera and Assigned(OnEnableControls) then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist SearchComplete OnEnableControls started');
    OnEnableControls(True);
    AddTextLog(CAMERA_GUID, 'Camlist SearchComplete OnEnableControls finished');
  end;
  if Status = 0 then
  begin
    AddTextLog(CAMERA_GUID,
      'Camlist SearchComplete SetStatus(RVCamera, STS_PRESENT);');
    SetStatus(RVCamera, STS_PRESENT);
    AddTextLog(CAMERA_GUID, 'Camlist SearchComplete RunCamera(RVCamera);');
    RunCamera(RVCamera);
  end
  else
  begin
    AddTextLog(CAMERA_GUID,
      'Camlist SearchComplete SetStatus(RVCamera, STS_NOTFOUND);');
    SetStatus(RVCamera, STS_NOTFOUND);
  end;
  AddTextLog(CAMERA_GUID, 'Camlist SearchComplete finished');
end;

procedure TCamListLoader.DoOnEndVideoStream(Sender: TObject; Status: Integer);
begin
  AddTextLog(CAMERA_GUID, 'Camlist DoOnEndVideoStream started');
  If Status = 0 then
    SetStatus(Sender as TRVCamera, STS_DONEVIDEO)
  else
    SetStatus(Sender as TRVCamera, STS_ERRORVIDEO);
  if Assigned(OnEndVideoStream) then
    OnEndVideoStream(Sender, Status);
  AddTextLog(CAMERA_GUID, 'Camlist DoOnEndVideoStream finsihed');
end;

procedure TCamListLoader.DoOnVideoStart(Sender: TObject);
begin
  AddTextLog(CAMERA_GUID, 'Camlist DoOnVideoStart started');
  SetStatus(Sender as TRVCamera, STS_PLAY);
  if Assigned(OnVideoStart) then
    OnVideoStart(Sender);
  AddTextLog(CAMERA_GUID, 'Camlist DoOnVideoStart finsihed');
end;

procedure TCamListLoader.DoOnStartVideoStream(Sender: TObject);
begin
  AddTextLog(CAMERA_GUID, 'Camlist DoOnStartVideoStream started');
  SetStatus(Sender as TRVCamera, STS_PLAYING);
  if Assigned(OnStartVideoStream) then
    OnStartVideoStream(Sender);
  AddTextLog(CAMERA_GUID, 'Camlist DoOnStartVideoStream finsihed');
end;

procedure TCamListLoader.SetStatus(Camera: TRVCamera; Status: string);
begin
  if Assigned(OnStatusChange) then
  begin
    AddTextLog(CAMERA_GUID, 'Camlist SetStatus started ' + Status);
    OnStatusChange(Camera, Status);
    AddTextLog(CAMERA_GUID, 'Camlist SetStatus finsihed' + Status);
  end;
end;

initialization

CamListLoader := TCamListLoader.Create;

finalization

CamListLoader.Destroy;

end.
