unit ImgFilter;

{$IFDEF FPC}
{$MODE Delphi}
{$ENDIF}

interface

uses
  {$IFNDEF FPC}
  Windows,
  {$ELSE}
  Classes, LCLIntf, LCLType, LMessages,
  {$ENDIF}
  SysUtils, Graphics, MRVBitmap, MRVType;

procedure GBlur(theBitmap: TRVMBitmap; radius: double);
procedure Emboss(ABitmap: TRVMBitmap; AMount: Integer);
// procedure Sharpen(sbm, tbm: TRVMBitmap; alpha: Single);
procedure Sharpen(sbm, tbm: TBitmap; alpha: Single);
procedure Antialiasing(C: TCanvas; Rect: TRect; Percent: Integer);
procedure EdgeDetect(var ABitmap: TRVMBitmap; level: Integer);
procedure ClearNoise(var ABitmap: TRVMBitmap; level: Integer);
procedure BmpAccentuation(Bmp: TRVMBitmap; var FinalBmp: TRVMBitmap;
  Correction: Integer);

implementation

uses Math;

const
  MaxKernelSize = 100;

type
  TKernelSize = 1 .. MaxKernelSize;

  TKernel = record
    Size: TKernelSize;
    Weights: array [-MaxKernelSize .. MaxKernelSize] of Single;
  end;
  // using TKernel, we ignore Weights, except for Weights in the range -Size..Size.

  TRGBTriple4 = record
    rgbtBlue: Byte;
    rgbtGreen: Byte;
    rgbtRed: Byte;
    rgbtAlpha: Byte;
  end;

  PRGBTriple = ^TRGBTriple4;

  PRow = ^TRow;
  TRow = array [0 .. 1000000] of TRGBTriple4;

  PPRows = ^TPRows;
  TPRows = array [0 .. 1000000] of PRow;

  {------------------------------------------------------------------------------}
  // Filter: Emboss
procedure Emboss(ABitmap: TRVMBitmap; AMount: Integer);
var
  x, y, i: Integer;
  p1, p2:  PByteArray;
begin
  for i := 0 to AMount do
  begin
    for y := 0 to ABitmap.Height - 2 do
    begin
      p1 := ABitmap.ScanLine(y);
      p2 := ABitmap.ScanLine(y + 1);
      for x := 0 to ABitmap.Width - 1 do
      begin
        p1[x * 4] := (p1[x * 4] + (p2[(x + 3) * 4] xor $FF)) shr 1;
        p1[x * 4 + 1] := (p1[x * 4 + 1] + (p2[(x + 3) * 4 + 1] xor $FF)) shr 1;
        p1[x * 4 + 2] := (p1[x * 4 + 1] + (p2[(x + 3) * 4 + 1] xor $FF)) shr 1;
      end;
    end;
  end;
end;

procedure BmpAccentuation(Bmp: TRVMBitmap; var FinalBmp: TRVMBitmap;
  Correction: Integer);
var
  Filter: array [0 .. 8] of Integer; // matrice de 3 * 3 pixels
  NewR, NewG, NewB, i, PosX, PosY, mX, mY, dX, dY, Diviseur, nColor, w,
    h: Integer;
begin
  for i := 0 to High(Filter) do
    if i in [0, 2, 6, 8] then
      Filter[i] := -Correction
    else if i = 4 then
      Filter[i] := (Correction * 4) + 128
      // +128 permet une correction bien йtalйe
    else
      Filter[i] := 0;
  Diviseur := Filter[4] - (Correction * 4);

  w := Bmp.Width;
  h := Bmp.Height;
  FinalBmp.SetSize(w, h);

  for PosY := 0 to h - 1 do
    for PosX := 0 to w - 1 do
    begin
      NewR := 0;
      NewG := 0;
      NewB := 0;
      for dY := -1 to 1 do
        for dX := -1 to 1 do
        begin
          // position du pixel а traiter
          mY := PosY + dY;
          mX := PosX + dX;

          if (mY >= 1) and (mY <= h - 1) and (mX >= 1) and (mX <= w - 1) then
            nColor := Bmp.Data[mY * w + mX]
          else
            nColor := Bmp.Data[PosY * w + PosX];

          i := 4 + (dY * 3) + dX; // I peut varier de 0 а 8
          NewR := NewR + (nColor and $FF) * Filter[i];
          NewG := NewG + ((nColor shr 8) and $FF) * Filter[i];
          NewB := NewB + ((nColor shr 16) and $FF) * Filter[i];
        end;

      NewR := NewR div Diviseur;
      NewG := NewG div Diviseur;
      NewB := NewB div Diviseur;
      if NewR > 255 then
        NewR := 255
      else if NewR < 0 then
        NewR := 0;
      if NewG > 255 then
        NewG := 255
      else if NewG < 0 then
        NewG := 0;
      if NewB > 255 then
        NewB := 255
      else if NewB < 0 then
        NewB := 0;
      FinalBmp.Data[PosY * w + PosX] := $FF000000 or
        Cardinal(NewR or (NewG shl 8) or (NewB shl 16));
    end;
  FinalBmp.Modified := True;
end;

{------------------------------------------------------------------------------}
// Filter: Sharpen
procedure Sharpen(sbm, tbm: TBitmap; alpha: Single);
// to sharpen, alpha must be >1.
// pixelformat pf24bit
// sharpens sbm to tbm
{var
 i, j, k: integer;
 sr: array[0..2] of PByte;
 st: array[0..4] of pRGBTriple;
 tr: PByte;
 tt, p: pRGBTriple;
 beta: Single;
 inta, intb: integer;
 bmh, bmw: integer;
 re, gr, bl: integer;
 BytesPerScanline: integer;
}
begin
  {   //sharpening is blending of the current pixel
    //with the average of the surrounding ones,
   //but with a negative weight for the average
   Assert((sbm.Width > 2) and (sbm.Height > 2), 'Bitmap must be at least 3x3');
   Assert((alpha > 1) and (alpha < 6), 'Alpha must be >1 and <6');
   beta := (alpha - 1) / 5; //we assume alpha>1 and beta<1
   intb := round(beta * $10000);
   inta := round(alpha * $10000); //integer scaled alpha and beta
   sbm.PixelFormat := pf24bit;
   tbm.PixelFormat := pf24bit;
   tbm.Width := sbm.Width;
   tbm.Height := sbm.Height;
   bmw := sbm.Width - 2;
   bmh := sbm.Height - 2;
   BytesPerScanline := (((bmw + 2) * 24 + 31) and not 31) div 8;

   tr := tbm.Scanline[0];
   tt := pRGBTriple(tr);

   sr[0] := sbm.Scanline[0];
   st[0] := pRGBTriple(sr[0]);
   for j := 0 to bmw + 1 do
   begin
   tt^ := st[0]^;
   inc(tt); inc(st[0]); //first row unchanged
   end;

   sr[1] := PByte(integer(sr[0]) - BytesPerScanline);
   sr[2] := PByte(integer(sr[1]) - BytesPerScanline);
   for i := 1 to bmh do
   begin
   Dec(tr, BytesPerScanline);
   tt := pRGBTriple(tr);
   st[0] := pRGBTriple(integer(sr[0]) + 3); //top
   st[1] := pRGBTriple(sr[1]); //left
   st[2] := pRGBTriple(integer(sr[1]) + 3); //center
   st[3] := pRGBTriple(integer(sr[1]) + 6); //right
   st[4] := pRGBTriple(integer(sr[2]) + 3); //bottom
   tt^ := st[1]^; //1st col unchanged
   for j := 1 to bmw do
   begin
   //calcutate average weighted by -beta
   re := 0; gr := 0; bl := 0;
   for k := 0 to 4 do
   begin
   re := re + st[k]^.rgbtRed;
   gr := gr + st[k]^.rgbtGreen;
   bl := bl + st[k]^.rgbtBlue;
   inc(st[k]);
   end;
   re := (intb * re + $7FFF) shr 16;
   gr := (intb * gr + $7FFF) shr 16;
   bl := (intb * bl + $7FFF) shr 16;

   //add center pixel weighted by alpha
   p := pRGBTriple(st[1]); //after inc, st[1] is at center
   re := (inta * p^.rgbtRed + $7FFF) shr 16 - re;
   gr := (inta * p^.rgbtGreen + $7FFF) shr 16 - gr;
   bl := (inta * p^.rgbtBlue + $7FFF) shr 16 - bl;

   //clamp and move into target pixel
   inc(tt);
   if re < 0 then
   re := 0
   else
   if re > 255 then
   re := 255;
   if gr < 0 then
   gr := 0
   else
   if gr > 255 then
   gr := 255;
   if bl < 0 then
   bl := 0
   else
   if bl > 255 then
   bl := 255;
   //this looks stupid, but avoids function calls

   tt^.rgbtRed := re;
   tt^.rgbtGreen := gr;
   tt^.rgbtBlue := bl;
   end;
   inc(tt);
   inc(st[1]);
   tt^ := st[1]^; //Last col unchanged
   sr[0] := sr[1];
   sr[1] := sr[2];
   Dec(sr[2], BytesPerScanline);
   end;
   // copy last row
   Dec(tr, BytesPerScanline);
   tt := pRGBTriple(tr);
   st[1] := pRGBTriple(sr[1]);
   for j := 0 to bmw + 1 do
   begin
   tt^ := st[1]^;
   inc(tt); inc(st[1]);
   end;}
end;

{------------------------------------------------------------------------------}
{The parameter "percent" needs an integer between 0 and 100
 (include zero and 100). If "Percent" is 0, there will be no effect.
 If it's 100 there will be the strongest effect.}
procedure Antialiasing(C: TCanvas; Rect: TRect; Percent: Integer);
var
  l, p:                   Integer;
  R, G, B:                Integer;
  R1, R2, G1, G2, B1, B2: Byte;
begin
  with C do
  begin
    for l := Rect.Top to Rect.Bottom do
    begin
      for p := Rect.Left to Rect.Right do
      begin
        R1 := GetRValue(Pixels[p, l]);
        G1 := GetGValue(Pixels[p, l]);
        B1 := GetBValue(Pixels[p, l]);

        // Pixel links
        // Pixel left
        R2 := GetRValue(Pixels[p - 1, l]);
        G2 := GetGValue(Pixels[p - 1, l]);
        B2 := GetBValue(Pixels[p - 1, l]);

        if (R1 <> R2) or (G1 <> G2) or (B1 <> B2) then
        begin
          R := Round(R1 + (R2 - R1) * 50 / (Percent + 50));
          G := Round(G1 + (G2 - G1) * 50 / (Percent + 50));
          B := Round(B1 + (B2 - B1) * 50 / (Percent + 50));
          Pixels[p - 1, l] := RGB(R, G, B);
        end;

        // Pixel rechts
        // Pixel right
        R2 := GetRValue(Pixels[p + 1, l]);
        G2 := GetGValue(Pixels[p + 1, l]);
        B2 := GetBValue(Pixels[p + 1, l]);

        if (R1 <> R2) or (G1 <> G2) or (B1 <> B2) then
        begin
          R := Round(R1 + (R2 - R1) * 50 / (Percent + 50));
          G := Round(G1 + (G2 - G1) * 50 / (Percent + 50));
          B := Round(B1 + (B2 - B1) * 50 / (Percent + 50));
          Pixels[p + 1, l] := RGB(R, G, B);
        end;

        // Pixel oben
        // Pixel up
        R2 := GetRValue(Pixels[p, l - 1]);
        G2 := GetGValue(Pixels[p, l - 1]);
        B2 := GetBValue(Pixels[p, l - 1]);

        if (R1 <> R2) or (G1 <> G2) or (B1 <> B2) then
        begin
          R := Round(R1 + (R2 - R1) * 50 / (Percent + 50));
          G := Round(G1 + (G2 - G1) * 50 / (Percent + 50));
          B := Round(B1 + (B2 - B1) * 50 / (Percent + 50));
          Pixels[p, l - 1] := RGB(R, G, B);
        end;

        // Pixel unten
        // Pixel down
        R2 := GetRValue(Pixels[p, l + 1]);
        G2 := GetGValue(Pixels[p, l + 1]);
        B2 := GetBValue(Pixels[p, l + 1]);

        if (R1 <> R2) or (G1 <> G2) or (B1 <> B2) then
        begin
          R := Round(R1 + (R2 - R1) * 50 / (Percent + 50));
          G := Round(G1 + (G2 - G1) * 50 / (Percent + 50));
          B := Round(B1 + (B2 - B1) * 50 / (Percent + 50));
          Pixels[p, l + 1] := RGB(R, G, B);
        end;
      end;
    end;
  end;
end;

{------------------------------------------------------------------------------}
// Filter: Edge detection
procedure EdgeDetect(var ABitmap: TRVMBitmap; level: Integer);
var
  x, y, sum:     Integer;
  p, p0, p1, p2: Integer;
  newBitmap:     TRVMBitmap;

  function getAvgColor(i: Integer): Integer;
  begin
    Result := ABitmap.Data[i];
    Result := ((Result and $FF) + ((Result shr 8) and $FF) +
      ((Result shr 16) and $FF)) div 3;
  end;

begin
  newBitmap := TRVMBitmap.Create(ABitmap.Width, ABitmap.Height);

  for y := 1 to ABitmap.Height - 2 do
  begin
    p0 := ABitmap.Width * (y - 1);
    p1 := ABitmap.Width * (y);
    p2 := ABitmap.Width * (y + 1);

    p := newBitmap.Width * y;
    for x := 1 to ABitmap.Width - 2 do
    begin
      sum := getAvgColor(p0 + (x - 1)) + getAvgColor(p0 + x) * 3 +
        getAvgColor(p0 + (x + 1)) + getAvgColor(p1 + (x - 1)) * 3 -
        getAvgColor(p1 + x) * 16 + getAvgColor(p1 + (x + 1)) * 3 +
        getAvgColor(p2 + (x - 1)) + getAvgColor(p2 + x) * 3 +
        getAvgColor(p2 + (x + 1));
      sum := Min(Max(sum * level div 9, 0), 255);
      newBitmap.Data[p + x] := $FF000000 or
        Cardinal(sum or (sum shl 8) or (sum shl 16));
    end;
  end;
  ABitmap.Assign(newBitmap);
  newBitmap.Free;
end;

{------------------------------------------------------------------------------}
// Filter: Clear noise
procedure ClearNoise(var ABitmap: TRVMBitmap; level: Integer);
var
  x, y, sum:     Integer;
  p, p0, p1, p2: PByteArray;
  newBitmap:     TRVMBitmap;
begin
  newBitmap := TRVMBitmap.Create(ABitmap.Width, ABitmap.Height);

  for y := 1 to ABitmap.Height - 2 do
  begin
    p0 := ABitmap.ScanLine(y - 1);
    p1 := ABitmap.ScanLine(y);
    p2 := ABitmap.ScanLine(y + 1);

    p := newBitmap.ScanLine(y);
    for x := 1 to ABitmap.Width - 2 do
    begin
      if p1[x * 4] < level then
        sum := 0
      else if (p0[(x - 1) * 4] < level) or (p0[x * 4] < level) or
        (p0[(x + 1) * 4] < level) or (p1[(x - 1) * 4] < level) or
        (p1[(x + 1) * 4] < level) or (p2[(x - 1) * 4] < level) or
        (p2[x * 4] < level) or (p2[(x + 1) * 4] < level) then
        sum := 0
      else
        sum := p1[x * 4];
      p[x * 4] := sum;
      p[x * 4 + 1] := sum;
      p[x * 4 + 2] := sum;
    end;
  end;
  ABitmap.Assign(newBitmap);
  newBitmap.Free;
end;

{------------------------------------------------------------------------------}
procedure MakeGaussianKernel(var K: TKernel; radius: double;
  MaxData, DataGranularity: double);

var
  j:           Integer;
  temp, delta: double;
  KernelSize:  TKernelSize;
begin

  for j := Low(K.Weights) to High(K.Weights) do
  begin
    temp := j / radius;
    K.Weights[j] := exp(-temp * temp / 2);
  end;

  temp := 0;
  for j := Low(K.Weights) to High(K.Weights) do
    temp := temp + K.Weights[j];
  for j := Low(K.Weights) to High(K.Weights) do
    K.Weights[j] := K.Weights[j] / temp;

  KernelSize := MaxKernelSize;
  delta := DataGranularity / (2 * MaxData);
  temp := 0;
  while (temp < delta) and (KernelSize > 1) do
  begin
    temp := temp + 2 * K.Weights[KernelSize];
    dec(KernelSize);
  end;

  K.Size := KernelSize;

  temp := 0;
  for j := -K.Size to K.Size do
    temp := temp + K.Weights[j];
  for j := -K.Size to K.Size do
    K.Weights[j] := K.Weights[j] / temp;

end;

function TrimInt(Lower, Upper, theInteger: Integer): Integer;
begin

  if (theInteger <= Upper) and (theInteger >= Lower) then
    Result := theInteger
  else if theInteger > Upper then
    Result := Upper
  else
    Result := Lower;
end;

function TrimReal(Lower, Upper: Integer; x: double): Integer;
begin

  if (x < Upper) and (x >= Lower) then
    Result := trunc(x)
  else if x > Upper then
    Result := Upper
  else
    Result := Lower;
end;

procedure BlurRow(var theRow: array of TRGBTriple4; K: TKernel; p: PRow);
var
  j, n:       Integer;
  tr, tg, tb: double;

  w: double;
begin

  for j := 0 to High(theRow) do
  begin
    tb := 0;
    tg := 0;
    tr := 0;
    for n := -K.Size to K.Size do
    begin
      w := K.Weights[n];

      with theRow[TrimInt(0, High(theRow), j - n)] do
      begin
        tb := tb + w * rgbtBlue;
        tg := tg + w * rgbtGreen;
        tr := tr + w * rgbtRed;
      end;
    end;
    with p[j] do
    begin
      rgbtBlue := TrimReal(0, 255, tb);
      rgbtGreen := TrimReal(0, 255, tg);
      rgbtRed := TrimReal(0, 255, tr);
    end;
  end;

  Move(p[0], theRow[0], (High(theRow) + 1) * 4);
end;

procedure GBlur(theBitmap: TRVMBitmap; radius: double);
var
  Row, Col: Integer;
  theRows:  PPRows;
  K:        TKernel;
  ACol:     PRow;
  p:        PRow;
begin
  MakeGaussianKernel(K, radius, 255, 1);
  GetMem(theRows, theBitmap.Height * SizeOf(PRow));
  GetMem(ACol, theBitmap.Height * 4);

  for Row := 0 to theBitmap.Height - 1 do
    theRows[Row] := PRow(theBitmap.ScanLine(Row));

  p := AllocMem(theBitmap.Width * 4);
  for Row := 0 to theBitmap.Height - 1 do
    BlurRow(Slice(theRows[Row]^, theBitmap.Width), K, p);

  ReAllocMem(p, theBitmap.Height * 4);
  for Col := 0 to theBitmap.Width - 1 do
  begin
    for Row := 0 to theBitmap.Height - 1 do
      ACol[Row] := theRows[Row][Col];

    BlurRow(Slice(ACol^, theBitmap.Height), K, p);

    for Row := 0 to theBitmap.Height - 1 do
      theRows[Row][Col] := ACol[Row];
  end;

  FreeMem(theRows);
  FreeMem(ACol);
  ReAllocMem(p, 0);
end;

end.
